from collections.abc import Callable
import os
from typing import Optional
from warnings import warn

from . import actions
from . import comms
from . import types
from . import validation

class MEUtility(object):
    def __init__(
        self,
        comms_path: str, 
        driver: str = None, 
        ignore_terminal_valid: bool = False, 
        ignore_driver_valid: bool = False
    ):
        """
        Initializes an instance of the MEUtility class.

        Args:
            comms_path (str) : The path to the communications resource (ex: 192.168.1.20).
            driver (str) : The driver name to use (ex: pycomm3 or pylogix).  If not specified, will default
                the first one installed that can be found.
            ignore_terminal_valid (bool) : If True, ignore terminal validation checks.
            ignore_driver_valid (bool) : If True, ignore driver validation checks.
        """
        self.comms_path = comms_path
        self.driver = driver
        self.ignore_terminal_valid = ignore_terminal_valid
        self.ignore_driver_valid = ignore_driver_valid

    def download(
        self, 
        file_path: str, 
        delete_logs: bool = False,
        overwrite: bool = False,
        replace_comms: bool = False,
        remote_file_name: str = None,
        run_at_startup: bool = True,
        progress: Optional[Callable[[str, int, int], None]] = None, 
    ) -> types.MEResponse:
        """
        Downloads a *.MER file from the local device to the remote terminal.

        Args:
            file_path (str) : The local path to the *.MER file (ex: C:\\YourFolder\\YourProgram.MER)            
            delete_logs (bool) : If True, will configure the terminal to delete logs atstartup.  Defaults to false.
            overwrite (bool): If True, will allow the download to overwrite an existing *.MER file on the terminal
                with the same file name.  Defaults to false.
            replace_comms (bool) : If True, will replace the terminal's communications setup with the one
                from the *.MER file being downloaded.  Defaults to false.
            remote_file_name (str) : If provided, is used to specify a different remote filename on the
                terminal than the local filename specified as part of file_path.
            run_at_startup (bool) : If True, will also set this *.MER file to be run at terminal startup and
                reboot the terminal now.  Defaults to True.
        """
        self.delete_logs = delete_logs
        self.overwrite = overwrite
        self.replace_comms = replace_comms
        if remote_file_name is None:
            self.remote_file_name = os.path.basename(file_path)
        else:
            self.remote_file_name = remote_file_name
        self.run_at_startup = run_at_startup

        # Use default MER directory if one is not specified
        if not os.path.isfile(file_path):
            if os.path.sep not in file_path:
                base_path = "C:\\Users\\Public\\Documents\\RSView Enterprise\\ME\\Runtime"
                file_path = os.path.join(base_path, file_path)

        with comms.Driver(self.comms_path, self.driver) as cip:
            file = types.MEFile(self.remote_file_name, self.overwrite, False, file_path)

            # Validate device at this communications path is a terminal of known version.
            self.device = validation.get_terminal_info(cip)
            if not(validation.is_valid_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')
                
            # Validate that all starting conditions for downnload to terminal are good
            try:
                resp = validation.is_valid_download(cip, self.device, file)
                if resp:
                    self.device.log.append(f'Validated download for {file.name}.')
                else:
                    self.device.log.append(f'Failed to validate download.')
                    return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to validate download.')
                return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)

            # Perform *.MER download to terminal
            try:
                resp = actions.download_mer_file(cip, self.device, file, self.run_at_startup, self.replace_comms, self.delete_logs, progress)
                if not(resp):
                    self.device.log.append(f'Failed to download to terminal.')
                    return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to download to terminal.')
                return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.MEResponseStatus.SUCCESS)
    
    def flash_firmware(
        self, 
        firmware_image_path: str, 
        firmware_helper_path: str, 
        progress: Optional[Callable[[str, int, int], None]] = None
    ) -> types.MEResponse:
        """
        Flashes a firmware image to the remote terminal.

        Args:
            firmware_image_path (str) : The local path to the firmware image file (ex: C:\\YourFolder\\FirmwareUpgradeCard\\upgrade\\SC.IMG)
            firmware_helper_path (str) : The local path to the firmware helper file (ex: C:\\Program Files (x86)\\Rockwell Software\\RSView Enterprise\\FUWhelper6xX.dll)
            progress : Optional callback for progress indication.
        """
        # Use default RSView directory if one is not specified
        if not os.path.isfile(firmware_helper_path):
            if os.path.sep not in firmware_helper_path:
                base_path = "C:\\Program Files (x86)\\Rockwell Software\\RSView Enterprise"
                firmware_helper_path = os.path.join(base_path, firmware_helper_path)

        with comms.Driver(self.comms_path, self.driver) as cip:
            if (self.driver == comms.DRIVER_NAME_PYCOMM3) and comms.is_routed_path(self.comms_path):
                if (cip._const_timeout_ticks != b'\xFF'):
                    if self.ignore_driver_valid:
                        warn('Drive pycomm3 specified with bad TIMEOUT_TICKS value, but driver validation is set to IGNORE.')
                    else:
                        resp = f"""
                            Cannot flash firmware to routed path using pycomm3 due to TIMEOUT_TICKS default value {cip._const_timeout_ticks}.
                            This will cause failures during the firmware upgrade process and require a factory reset to recover.
                            Please change the value in pycomm3.const.TIMEOUT_TICKS to b'\\xFF' to proceed, or use pylogix instead.
                        """
                        raise NotImplementedError(resp)

            # Set socket timeout first.
            # The terminal will pause at certain points and delay acknowledging messages.
            # Without this, the process will fail and the terminal will require a factory reset.
            cip.timeout = 255.0

            # Validate device at this communications path is a terminal of known version.
            self.device = validation.get_terminal_info(cip)
            if not(validation.is_valid_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            # Perform firmware flash to terminal
            try:
                resp = actions.flash_firmware_upgrade_card(
                    cip=cip,
                    device=self.device,
                    firmware_image_path=firmware_image_path,
                    firmware_helper_path=firmware_helper_path,
                    progress=progress)

                if not(resp):
                    self.device.log.append(f'Failed to flash terminal.')
                    return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to flash terminal.')
                return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.MEResponseStatus.SUCCESS)

    def get_terminal_info(
        self, 
        print_log: bool = False,
        redact_log: bool = False,
        silent_mode: bool = False,
    ) -> types.MEResponse:
        """
        If no upload or download are desired, where terminal info would typically be checked
        as a prerequisite, this function can be called to generate similar log entries to
        get information about the remote terminal.

        Args:
            print_log (bool) : If True, will print values.  Please include
                a copy with any bug reports!  Defaults to false.
            redact_log (bool) : If True, will exclude potentially sensitive values
                such as the *.MER file names on the remote terminal. Defaults to False.
            silent_mode (bool) : If True, will exclude values that require invocations
                or transfers so that Diagnostics window doesn't pop up on the remote
                terminal.  Defaults to False.
        """
        self.print_log = print_log
        self.redact_log = redact_log
        self.silent_mode = silent_mode
        with comms.Driver(self.comms_path, self.driver) as cip:
            self.device = validation.get_terminal_info(cip)
            if not((validation.is_valid_me_terminal(self.device) or validation.is_valid_dmk_terminal(self.device))):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            try:
                actions.create_log(cip, self.device, self.print_log, self.redact_log, self.silent_mode)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to get terminal info.')
                return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.MEResponseStatus.SUCCESS)

    def reboot(
        self
    ) -> types.MEResponse:
        """
        Reboots the remote terminal now.
        """
        with comms.Driver(self.comms_path, self.driver) as cip:
            self.device = validation.get_terminal_info(cip)

            if not(validation.is_valid_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            try:
                actions.reboot(cip, self.device)
            except Exception as e:
                self.device.log.append(f'Failed to reboot terminal.')
                return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.MEResponseStatus.SUCCESS)

    def upload(
        self, 
        file_path: str, 
        overwrite: bool = False,
        remote_file_name: str = None,
        progress: Optional[Callable[[str, int, int], None]] = None, 
    ) -> types.MEResponse:
        """
        Uploads a *.MER file from the remote terminal to the local device.

        Args:
            file_path (str) : The local path to the *.MER file (ex: C:\\YourFolder\\YourProgram.MER)
            overwrite (bool) : If True, will replace the file on the local device with the uploaded
                copy from the remote terminal.  Defaults to False.
            remote_file_name (str) : If provided, is used to specify a different remote filename on
                the terminal than the local filename specified as part of file_path where it will end up.
        """
        file = types.MEFile(os.path.basename(file_path), False, False, file_path)
        if (remote_file_name is None):
            self.remote_file_name = file.name
        else:
            self.remote_file_name = remote_file_name
        self.overwrite = overwrite
        rem_file = types.MEFile(self.remote_file_name,False,False,file_path)

        # Create upload folder if it doesn't exist yet
        if not(os.path.exists(os.path.dirname(file.path))): os.makedirs(os.path.dirname(file.path))

        with comms.Driver(self.comms_path, self.driver) as cip:
            # Validate device at this communications path is a terminal of known version.
            self.device = validation.get_terminal_info(cip)
            if not(validation.is_valid_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            # Check for existing *.MER
            if not(self.overwrite) and (os.path.exists(file.path)):
                self.device.log.append(f'File {file.path} already exists.  Use kwarg overwrite=True to overwrite existing local file from the remote terminal.')
                return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)

            # Perform *.MER upload from terminal
            try:
                resp = actions.upload_mer_file(cip, self.device, file, rem_file, progress)
                if not(resp):
                    self.device.log.append(f'Failed to upload from terminal.')
                    return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to upload from terminal.')
                return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.MEResponseStatus.SUCCESS)

    def upload_all(
        self, 
        file_path: str, 
        overwrite: bool = False,
        progress: Optional[Callable[[str, int, int], None]] = None, 
    ) -> types.MEResponse:
        """
        Uploads all *.MER files from the remote terminal to the local device.

        Args:
            file_path (str) : The local path to the target directory (ex: C:\\YourFolder)
            overwrite (bool) : If True, will replace the file on the local device with the 
                uploaded copy from the remote terminal.  Defaults to False.
        """
        self.overwrite = overwrite

        # Create upload folder if it doesn't exist yet
        if not(os.path.exists(file_path)): os.makedirs(os.path.dirname(file_path))

        with comms.Driver(self.comms_path, self.driver) as cip:
            # Validate device at this communications path is a terminal of known version.
            self.device = validation.get_terminal_info(cip)
            if not(validation.is_valid_me_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            try:
                mer_list = actions.upload_mer_list(cip, self.device)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to upload *.MER list from terminal.')
                return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)
            
            for mer in mer_list:
                if len(mer) > 0:
                    mer_path = os.path.join(file_path, mer)
                    file = types.MEFile(os.path.basename(mer_path), self.overwrite, False, mer_path)

                    # Check for existing *.MER
                    if not(self.overwrite) and (os.path.exists(mer_path)):
                        self.device.log.append(f'File {mer_path} already exists.  Use kwarg overwrite=True to overwrite existing local file from the remote terminal.')
                        return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)
                    
                    # Perform *.MER upload from terminal
                    try:
                        resp = actions.upload_mer_file(cip, self.device, file, file, progress)
                        if not(resp):
                            self.device.log.append(f'Failed to upload from terminal.')
                            return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)
                    except Exception as e:
                        self.device.log.append(f'Exception: {str(e)}')
                        self.device.log.append(f'Failed to upload from terminal.')
                        return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.MEResponseStatus.SUCCESS)


from collections.abc import Callable
import configparser
import struct
import time
from typing import Optional
from warnings import warn
import zipfile

from . import comms
from . import messages
from . import types

FILE_NANE_CONTENT = 'Content.txt'
FILE_NAME_NVS = 'RA_PVPApps_FTviewME_AllRegions.nvs'

def deserialize_dmk_content_header(config: configparser) -> types.DMKContentHeader:
    section_name = 'Content'
    if section_name not in config:
        raise ValueError(f"Section '{section_name}' not found in configuration.")

    section = config[section_name]
    return types.DMKContentHeader(
        dmk_type=section.get('DMKType', ''),
        family_name=section.get('FamilyName',''),
        revision=section.get('Revision',''),
        number_catalogs=section.getint('NumberCatalogs',0)
    )

def deserialize_dmk_content_catalog(config: configparser, catalog_num: int) -> types.DMKContentCatalog:
    section_name = f'Catalog{catalog_num}'
    if section_name not in config:
        raise ValueError(f"Section '{section_name}' not found in configuration.")

    section = config[section_name]
    return types.DMKContentCatalog(
        catalog=section.get('Catalog',''),
        revision=section.get('Revision',''),
        vendor_id=int(section.get('VendorId',0),16),
        vendor_id_mask=int(section.get('VendorIdMask',0),16),
        product_type=int(section.get('ProductType',0),16),
        product_type_mask=int(section.get('ProductTypeMask',0),16),
        product_code=int(section.get('ProductCode',0),16),
        product_code_mask=int(section.get('ProductCodeMask',0),16),
        device_instance=section.getint('DeviceInstance',0),
        disable_sub_minor=section.getint('DisableSubminor',0),
        driver_name=section.get('DriverName','')
    )

def deserialize_dmk_content_file(config: configparser) -> types.DMKContentFile:
    header = deserialize_dmk_content_header(config)
    catalogs = []
    for i in range(header.number_catalogs):
        catalogs.append(deserialize_dmk_content_catalog(config, i+1))

    return types.DMKContentFile(
        header=header,
        catalogs=catalogs
    )

def deserialize_dmk_nvs_header(config: configparser) -> types.DMKNvsHeader:
    section_name = 'Device'
    if section_name not in config:
        raise ValueError(f"Section '{section_name}' not found in configuration.")

    section = config[section_name]
    return types.DMKNvsHeader(
        description=section.get('Description',''),
        programming_protocol=section.get('ProgrammingProtocol',''),
        disable_sub_minor=section.getint('DisableSubminor',0),
        new_revision=section.get('NewRevision',''),
        number_updates=section.getint('NumberUpdates',0),
        connection_type=section.get('ConnectionType',''),
        max_power_up_seconds=section.getint('MaxPowerupSeconds',0),
        number_identities=section.getint('NumberIdentities',0),
        family_name=section.get('FamilyName',''),
        release_notes_file_name=section.get('ReleaseNotesFileName','')
    )

def deserialize_dmk_nvs_update(config: configparser, update_num: int) -> types.DMKNvsUpdate:
    section_name = f'Update{update_num}'
    if section_name not in config:
        raise ValueError(f"Section '{section_name}' not found in configuration.")

    section = config[section_name]
    return types.DMKNvsUpdate(
        nvs_instance=section.getint('NVSInstance',0),
        major_revision=section.getint('MajorRevision',0),
        minor_revision=section.getint('MinorRevision',0),
        tftp_timeout_seconds=section.getint('TFTPTimeOutSeconds',0),
        max_timeout_seconds=section.getint('MaxTimeoutSeconds',0),
        starting_location=int(section.get('StartingLocation',0),16),
        file_size=section.getint('FileSize',0),
        data_file_name=section.get('DataFileName',''),
        rename_data_file=section.get('RenameDataFile',''),
        update_reset=section.getint('UpdateReset',0),
        auto_reset_on_error=section.getint('AutoResetOnError',0),
        first_transfer_delay=section.getint('FirstTransferDelay',0),
        last_transfer_delay=section.getint('LastTransferDelay',0),
        error_instructions=section.get('ErrorInstructions')
    )

def deserialize_dmk_nvs_file(config: configparser) -> types.DMKNvsFile:
    header = deserialize_dmk_nvs_header(config)
    updates = []
    for i in range(header.number_updates):
        updates.append(deserialize_dmk_nvs_update(config, i+1))

    return types.DMKNvsFile(
        header=header,
        updates=updates
    )

def send_dmk_update_preamble(cip: comms.Driver, instance: int, serial_number: str, file_size: int) -> int:
    """
    Sends a DMK CAB file preamble message to the remote terminal.
    It responds with the chunk size to use for the actual file
    transfer.

    Args:
        cip (comms.Driver): CIP driver to communicate with the terminal
        instance (int): File instance
        serial_number (str): The terminal serial number as an 8-character hex string 'FFFFFFFF'
        file_size (int): File size in byte to that will be transferred

    Returns:
        int: The chunk size to use with DMK file transfer.

    Raises:
        Exception: If the file transfer fails or if the response contains
        unexpected values, indicating potential issues with the transfer.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | File size in bytes                                  |
        | Bytes 4->7    | Unknown purpose (reserved?)                         |
        | Bytes 8->11   | Unknown purpose                                     |
        | Bytes 12->15  | Target device serial number                         |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Unknown purpose                                     |
        | Bytes 4->7    | Chunk size in bytes                                 |
        | Bytes 8->11   | Unknown purpose                                     |
    """
    warn('Preamble using static UNK1 values')
    req_unk1 = 0
    req_unk2 = 0x00010007
    req_serial_number = int(serial_number, 16)
    req_data = struct.pack('<IIII', file_size, req_unk1, req_unk2, req_serial_number)

    resp = messages.dmk_preamble(cip, instance, req_data)

    if not resp.value: raise Exception(f'Failed to write file preamble to terminal.')
    resp_unk1, resp_chunk_size, resp_unk2 = struct.unpack('<III', resp.value)
    if (resp_unk1 != 0): raise Exception(f'Response UNK1: {resp_unk1}.  Update failed.')
    if (resp_chunk_size <= 0): raise Exception(f'Response chunk size: {resp_chunk_size}.  Update failed.')
    if (resp_unk2 != 0): raise Exception(f'Response UNK2: {resp_unk2}.  Update failed.')
    return resp_chunk_size

def send_dmk_update_file(cip: comms.Driver, instance: int, chunk_size: int, source_data: bytearray, description: str, progress: Optional[Callable[[str, int, int], None]] = None):
    req_offset = 0
    current_bytes = 0
    total_bytes = len(source_data)
    if progress: progress(f'{description}',total_bytes, current_bytes)
    while current_bytes < total_bytes:
        req_chunk = source_data[req_offset:req_offset + chunk_size]
        req_header = struct.pack('<I', req_offset)
        req_data = req_header + req_chunk

        resp = messages.dmk_chunk(cip, instance, req_data)

        if not resp: raise Exception(f'Failed to write chunk to terminal.')
        resp_offset, resp_offset_next, resp_code = struct.unpack('<IIH', resp.value)
        if (resp_offset != req_offset): raise Exception(f'Response offset: {resp_offset} does not match request offset: {req_offset}.  Update failed.')
        if (resp_code != 0x02): raise Exception(f'Response code: {resp_code}.  Update failed.')

        if (resp_offset_next == 0xFFFFFFFF):
            # End of file
            current_bytes = total_bytes
        else:
            # Next chunk
            current_bytes = resp_offset_next

        # Update progress callback
        if progress: progress(f'{description}',total_bytes, current_bytes)
            
        # Continue to next chunk
        req_offset = resp_offset_next
        #req_offset += len(req_chunk)

def send_dmk_reset(cip: comms.Driver):
    """
    Sends a reset after DMK is transferred to the rmote terminal..

    Args:
        cip (comms.Driver): CIP driver to communicate with the terminal
    """
    req_data = struct.pack('<B', 0x00)
    resp = messages.reset(cip, req_data)
    if not resp: raise Exception(f'Failed to reset terminal.')

def send_dmk_updates(cip: comms.Driver, device: types.MEDeviceInfo, dmk_file_path: str, nvs: types.DMKNvsFile, progress: Optional[Callable[[str, int, int], None]] = None):
    cip.close()
    with zipfile.ZipFile(dmk_file_path, 'r') as zf:
        instance = 0
        cip.connection_size = 1400
        for update in nvs.updates:
            instance += 1
            cip.forward_open()
            cip.sequence_reset()

            chunk_size = send_dmk_update_preamble(
                cip=cip,
                instance=instance,
                serial_number=device.cip_identity.serial_number,
                file_size=update.file_size
            )
            with zf.open(update.data_file_name) as file:
                send_dmk_update_file(
                    cip=cip,
                    instance=instance,
                    chunk_size=chunk_size,
                    source_data=bytearray(file.read()),
                    description=f'Updating {instance} of {len(nvs.updates)} {file.name}',
                    progress=progress
                )
            cip.forward_close()
            if update.update_reset: send_dmk_reset(cip)

def validate_update_size(dmk_file_path: str, nvs: types.DMKNvsFile):
    with zipfile.ZipFile(dmk_file_path, 'r') as zf:
        for update in nvs.updates:
            actual_size = zf.getinfo(update.data_file_name).file_size
            with zf.open(update.data_file_name) as file:
                if (actual_size != update.file_size):
                    raise Exception(f'File: {update.data_file_name}, Expected Size: {update.file_size}, Actual Size: {actual_size}')

def process_dmk(cip: comms.Driver, device: types.MEDeviceInfo, dmk_file_path: str, dry_run: bool, progress: Optional[Callable[[str, int, int], None]] = None):
    config_content = configparser.ConfigParser(allow_unnamed_section=True)
    config_nvs = configparser.ConfigParser(allow_unnamed_section=True, allow_no_value=True)
    with zipfile.ZipFile(dmk_file_path, 'r') as zf:
        with zf.open(FILE_NANE_CONTENT) as file:
            raw_file = file.read().decode('utf-8')
            config_content.read_string(raw_file)
        with zf.open(FILE_NAME_NVS) as file:
            raw_file = file.read().decode('utf-8', errors='ignore')
            config_nvs.read_string(raw_file)

    dmk_file = types.DMKFile(
        content=deserialize_dmk_content_file(config_content),
        nvs=deserialize_dmk_nvs_file(config_nvs)
    )
    validate_update_size(dmk_file_path, dmk_file.nvs)
    if not(dry_run):
        send_dmk_updates(
            cip=cip,
            device=device,
            dmk_file_path=dmk_file_path,
            nvs=dmk_file.nvs,
            progress=progress
        )
    return dmk_file

def masked_equals(mask: int, a: int, b: int) -> bool:
    return (mask & a) == (mask & b)

def validate_dmk_for_terminal(device: types.MEDeviceInfo, content: types.DMKContentFile) -> bool:
    for catalog in content.catalogs:
        if not(masked_equals(catalog.vendor_id_mask, catalog.vendor_id, device.cip_identity.vendor_id)): continue
        if not(masked_equals(catalog.product_type_mask, catalog.product_type, device.cip_identity.product_type)): continue
        if not(masked_equals(catalog.product_code_mask, catalog.product_code, device.cip_identity.product_code)): continue
        return True # Device matches this catalog
    return False # Device doesn't match any catalogs


from enum import StrEnum

from . import helper
from .. import comms
from .. import types

# Known static value for successful creation of a folder.
# Further investigation needed.
CREATE_DIR_SUCCESS = 0

# Known functions available from FuwHelper.
class FuwHelperFunctions(StrEnum):
    CREATE_FOLDER = 'CreateRemDirectory' # Args: {Folder Path}, Returns: Static value if successful [Further investigation needed]
    DELETE_FILE = 'DeleteRemFile' # Args: {File Path}, Returns: ???
    GET_FILE_EXISTS = 'FileExists' # Args: {File Path}, Returns: {File Size} if {File Path} exists, 0 otherwise
    GET_FOLDER_EXISTS = 'StorageExists' # Args: {Folder Path}, Returns: 1 if {Folder Path} exists
    GET_PROCESS_RUNNING = 'IsExeRunning' # Args: {Process Name}, Resp Code: 0 successful, Resp Data: 1 if process is running, 0 otherwise
    STOP_PROCESS_ME = 'SafeTerminateME' # Args: Null, Resp Code: 0 successful, 2 failed/wasn't running, Resp Data: Null

def create_folder(cip: comms.Driver, paths: types.MEPaths, dir: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.CREATE_FOLDER, dir]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != CREATE_DIR_SUCCESS): raise Exception(f'Failed to execute function: {req_args}, response code: {resp_code}, response data: {resp_data}.')
    return True

def delete_file(cip: comms.Driver, paths: types.MEPaths, file_path: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.DELETE_FILE, file_path]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to delete file on terminal: {file_path}, response code: {resp_code}, response data: {resp_data}.')
    return True

def get_process_running(cip: comms.Driver, paths: types.MEPaths, process_name: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.GET_PROCESS_RUNNING, process_name]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): return False    
    return bool(int(resp_data))

def get_file_exists(cip: comms.Driver, paths: types.MEPaths, file_path: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.GET_FILE_EXISTS, file_path]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): return False    
    return bool(int(resp_data))

def get_folder_exists(cip: comms.Driver, paths: types.MEPaths, folder_path: str) -> bool:
    req_args = [paths.fuwhelper_file, FuwHelperFunctions.GET_FOLDER_EXISTS, folder_path]
    resp_code, resp_data = helper.run_function(cip, req_args)
    if (resp_code != 0): return False    
    return bool(int(resp_data))

def stop_process_me(cip: comms.Driver, paths: types.MEPaths) -> bool:
    if get_process_running(cip, paths, 'MERuntime.exe'):
        req_args = [paths.fuwhelper_file, FuwHelperFunctions.STOP_PROCESS_ME, '']
        resp_code, resp_data = helper.run_function(cip, req_args)
        if (resp_code != 0): return False
    return True



from collections.abc import Callable
import os
from typing import Optional
from warnings import warn

from . import comms
from . import dmk
from . import types
from . import validation

class CFUtility(object):
    def __init__(
        self,
        comms_path: str, 
        driver: str = None, 
        ignore_terminal_valid: bool = False, 
        ignore_driver_valid: bool = False
    ):
        """
        Initializes an instance of the MEUtility class.

        Args:
            comms_path (str) : The path to the communications resource (ex: 192.168.1.20).
            driver (str) : The driver name to use (ex: pycomm3 or pylogix).  If not specified, will default
                the first one installed that can be found.
            ignore_terminal_valid (bool) : If True, ignore terminal validation checks.
            ignore_driver_valid (bool) : If True, ignore driver validation checks.
        """
        self.comms_path = comms_path
        self.driver = driver
        self.ignore_terminal_valid = ignore_terminal_valid
        self.ignore_driver_valid = ignore_driver_valid

    def flash_firmware(
        self, 
        firmware_image_path: str, 
        dry_run: bool = False,
        progress: Optional[Callable[[str, int, int], None]] = None
    ) -> types.MEResponse:
        """
        Flashes a firmware image to the remote terminal.

        Args:
            firmware_image_path (str) : The local path to the firmware image file (ex: C:\\YourFolder\\\\FirmwareImage.DMK)
            dry_run (bool) : If True, run through pre-processing but don't actually flash the device.
            progress: Optional callback for progress indication.
        """

        with comms.Driver(self.comms_path, self.driver) as cip:
            if (self.driver == comms.DRIVER_NAME_PYLOGIX):
                if self.ignore_driver_valid:
                    warn('Drive pylogix specified but driver validation is set to IGNORE.')
                else:
                    resp = f"""
                        Cannot flash firmware with pylogix yet, work still required on forward open/forward close.
                        Please use pycomm3 instead for now.
                    """
                    raise NotImplementedError(resp)


            # Set socket timeout first.
            # The terminal will pause at certain points and delay acknowledging messages.
            # Without this, the process will fail and the terminal will require a factory reset.
            cip.timeout = 255.0

            # Validate device at this communications path is a terminal of known version.
            self.device = validation.get_terminal_info(cip)
            if not(validation.is_valid_dmk_terminal(self.device)):
                if self.ignore_terminal_valid:
                    warn('Invalid device selected, but terminal validation is set to IGNORE.')
                else:
                    raise Exception('Invalid device selected.  Use kwarg ignore_terminal_valid=True when initializing MEUtility object to proceed at your own risk.')

            # Perform firmware flash to terminal
            try:
                resp = dmk.process_dmk(
                    cip=cip,
                    device=self.device,
                    dmk_file_path=firmware_image_path,
                    dry_run=dry_run,
                    progress=progress)                    
                if not(resp):
                    self.device.log.append(f'Failed to flash terminal.')
                    return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)
            except Exception as e:
                self.device.log.append(f'Exception: {str(e)}')
                self.device.log.append(f'Failed to flash terminal.')
                return types.MEResponse(self.device, types.MEResponseStatus.FAILURE)

        return types.MEResponse(self.device, types.MEResponseStatus.SUCCESS)


from enum import IntEnum
from . import comms

class DMKClasses(IntEnum):
    FIRMWARE_FLASH = int.from_bytes(b'\xA1', byteorder='big')

class DMKServices(IntEnum):
    FIRMWARE_PREAMBLE = int.from_bytes(b'\x51', byteorder='big')
    FIRMWARE_CHUNK = int.from_bytes(b'\x52', byteorder='big')

class MEClasses(IntEnum):
    FUNCTION = int.from_bytes(b'\x04\xFD', byteorder='big')
    REGISTRY = int.from_bytes(b'\x04\xFE', byteorder='big')
    FILE = int.from_bytes(b'\x04\xFF', byteorder='big')

class MEServices(IntEnum):
    CREATE = int.from_bytes(b'\x08', byteorder='big')
    DELETE = int.from_bytes(b'\x09', byteorder='big')
    GET_ATTRIBUTE_SINGLE = int.from_bytes(b'\x0e', byteorder='big')
    SET_ATTRIBUTE_SINGLE = int.from_bytes(b'\x10', byteorder='big')
    EXECUTE = int.from_bytes(b'\x50', byteorder='big')
    READ_REGISTRY = int.from_bytes(b'\x51', byteorder='big')
    WRITE_FILE = int.from_bytes(b'\x52', byteorder='big')
    READ_FILE = int.from_bytes(b'\x53', byteorder='big')

def create_transfer_instance(cip: comms.Driver, data):
    return cip.generic_message(
        service=MEServices.CREATE,
        class_code=MEClasses.FILE,
        instance=0x00,
        attribute=None,
        request_data=data
    )

def delete_transfer_instance(cip: comms.Driver, transfer_instance: int):
    return cip.generic_message(
        service=MEServices.DELETE,
        class_code=MEClasses.FILE,
        instance=transfer_instance,
        attribute=None
    )

def dmk_preamble(cip: comms.Driver, instance: int, data):
    return cip.generic_message(
        service=DMKServices.FIRMWARE_PREAMBLE,
        class_code=DMKClasses.FIRMWARE_FLASH,
        instance=instance,
        attribute=None,
        request_data=data,
        connected=True
    )

def dmk_chunk(cip: comms.Driver, instance: int, data):
    return cip.generic_message(
        service=DMKServices.FIRMWARE_CHUNK,
        class_code=DMKClasses.FIRMWARE_FLASH,
        instance=instance,
        attribute=None,
        request_data=data,
        connected=True
    )

def get_identity(cip: comms.Driver):
    return cip.generic_message(
        service=0x01,
        class_code=0x01,
        instance=0x01,
        attribute=None
    )

def get_hardware_rev(cip: comms.Driver):
    return cip.generic_message(
        service=MEServices.GET_ATTRIBUTE_SINGLE,
        class_code=0x01,
        instance=0x01,
        attribute=101,
    )

def reset(cip: comms.Driver, data):
    return cip.generic_message(
        service=0x05,
        class_code=0x01,
        instance=0x01,
        attribute=None,
        request_data=data
    )

def read_file_chunk(cip: comms.Driver, transfer_instance: int, data):
    return cip.generic_message(
        service=MEServices.READ_FILE,
        class_code=MEClasses.FILE,
        instance=transfer_instance,
        attribute=None,
        request_data=data
    )

def read_registry(cip: comms.Driver, data):
    return cip.generic_message(
        service=MEServices.READ_REGISTRY,
        class_code=MEClasses.REGISTRY,
        instance=0x00,
        attribute=None,
        request_data=data
    )

def run_function(cip: comms.Driver, data):
    return cip.generic_message(
        service=MEServices.EXECUTE,
        class_code=MEClasses.FUNCTION,
        instance=0x00,
        attribute=None,
        request_data=data
    )

def write_file_chunk(cip: comms.Driver, transfer_instance: int, data):
    return cip.generic_message(
        service=MEServices.WRITE_FILE,
        class_code=MEClasses.FILE,
        instance=transfer_instance,
        attribute=None,
        request_data=data
    )

def get_attr_unk(cip: comms.Driver, data):
    return cip.generic_message(
        service=MEServices.GET_ATTRIBUTE_SINGLE,
        class_code=MEClasses.FILE,
        instance=0x00,
        attribute=None,
        request_data=data
    )

def set_attr_unk(cip: comms.Driver, data):
    return cip.generic_message(
        service=MEServices.SET_ATTRIBUTE_SINGLE,
        class_code=MEClasses.FILE,
        instance=0x01,
        attribute=None,
        request_data=data
    )


from collections.abc import Callable
import os
import time
import traceback
from typing import Optional

from .terminal import files
from .terminal import fuwhelper
from .terminal import helper
from .terminal import registry
from . import comms
from . import types

def create_log(cip: comms.Driver, device: types.MEDeviceInfo, print_log: bool, redact_log: bool, silent_mode: bool):
    if print_log: print(f'Terminal CIP identity: {device.cip_identity}.')
    if print_log: print(f'Terminal ME identity: {device.me_identity}.')
    if print_log: print(f'Terminal ME paths: {device.me_paths}.')

    try:
        line = f'Terminal has {helper.get_free_space_runtime(cip, device.me_paths)} free bytes.'
    except:
        line = f'Failed to get free space on terminal.'
    device.log.append(line)
    if print_log: print(f'{line}')

    if not silent_mode:
        try:
            files = upload_med_list(cip, device)
            if redact_log: files = ['Redacted' for _ in files]
            line = f'Terminal has MED files: {files}.'
            if len(files) > 0: device.running_med_file = files[0]
        except:
            line = f'Failed to list MED files on terminal.'
        device.log.append(line)
        if print_log: print(f'{line}')

        try:
            files = upload_mer_list(cip, device)
            if redact_log: files = ['Redacted' for _ in files]
            line = f'Terminal has MER files: {files}.'
        except:
            line = f'Failed to list MER files on terminal.'
        device.log.append(line)
        if print_log: print(f'{line}')

    try:
        file = registry.get_startup_mer(cip)
        if redact_log: file = 'Redacted'
        line = f'Terminal startup file: {file}.'
        if file.lower().endswith('.mer'): device.startup_mer_file = file.split('\\')[-1]
    except:
        if device.cip_identity.major_rev < 6:
            # For PanelView Plus 5.10 and earlier this registry key appears to be unavailable.
            line = f'Terminal startup file: could not be determined due to hardware version.'
        else:
            # If no startup app has been defined, it will also fail. 
            line = f'Terminal startup file: not configured.'
    device.log.append(line)
    if print_log: print(f'{line}')

def download(cip: comms.Driver, device: types.MEDeviceInfo, file: types.MEFile, rem_path: str, file_data: bytearray, progress: Optional[Callable[[str, int, int], None]] = None) -> bool:
    # Get attributes
    #
    # Still no clue on what these are, or when/how they would change.
    # If they aren't changed by creating paths, could be moved ahead
    # to is_download_valid().
    try:
        get_unk1 = files.is_get_unk_valid(cip)
        if get_unk1:
            device.log.append(f'Got UNK1 attributes from terminal.')
        else:
            device.log.append(f'Got invalid UNK1 attributes from terminal.  Please file a bug report with all available information.')
            return False
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to get UNK1 attributes from terminal.')
        return False

    # Create a transfer instance on the terminal
    try:
        transfer_instance = files.create_transfer_instance_download(cip, file, rem_path)
        device.log.append(f'Create transfer instance {transfer_instance} for download.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create transfer instance for download')
        return False

    # Set attributes
    #
    # Still no clue what this is.  Might be setting file up for write?
    continue_download = True
    try:
        set_unk1 = files.is_set_unk_valid(cip)
        if set_unk1:
            device.log.append(f'Set UNK1 attributes on terminal.')
        else:
            device.log.append(f'Failed to set UNK1 attributes on terminal.  Please file a bug report with all available information.')
            continue_download = False
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to set UNK1 attributes on terminal.')
        continue_download = False

    # Transfer file chunk by chunk
    try:
        if continue_download:
            files.execute_transfer_download(cip, transfer_instance, file_data, progress)
            device.log.append(f'Downloaded {file.name} using transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to download {file.name} using transfer instance {transfer_instance}.')

    # Delete transfer instance on the terminal
    try:
        files.delete_transfer_instance(cip, transfer_instance)
        device.log.append(f'Deleted transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to delete transfer instance {transfer_instance}.')

    return continue_download

def download_file(cip: comms.Driver, device: types.MEDeviceInfo, file: types.MEFile, rem_path: str, progress: Optional[Callable[[str, int, int], None]] = None) -> bool:
    with open(file.path, 'rb') as source_file:
        return download(cip, device, file, rem_path, bytearray(source_file.read()), progress)

def download_mer_file(cip: comms.Driver, device: types.MEDeviceInfo, file:types.MEFile, run_at_startup: bool, replace_comms: bool, delete_logs: bool, progress: Optional[Callable[[str, int, int], None]] = None) -> bool:
    # Create runtime folder
    try:
        helper.create_folder_runtime(cip, device.me_paths)
        device.log.append(f'Create runtime directory on terminal.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create runtime directory on terminal.')
        return False

    # Perform download
    try:
        continue_download = download_file(cip, device, file, device.me_paths.runtime, progress)
        device.log.append(f'Downloaded {file.path} to {device.me_paths.runtime}.')
    except:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to download {file.path} to {device.me_paths.runtime}.')
        return False

    # Set *.MER to run at startup and then reboot
    try:
        if continue_download:
            if run_at_startup:
                helper.create_me_shortcut(cip, device.me_paths, file.name, replace_comms, delete_logs)
                device.log.append(f'Set file: {file.name} to run at startup with Replace Comms: {replace_comms}, Delete Logs: {delete_logs}.')
                reboot(cip, device)
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to set file to run at startup.')
        
    return continue_download

def flash_firmware_upgrade_card(cip: comms.Driver, 
                                device: types.MEDeviceInfo, 
                                firmware_image_path: str,
                                firmware_helper_path: str, 
                                progress: Optional[Callable[[str, int, int], None]] = None
                                ):

    # Determine if firmware upgrade helepr exists already
    transfer_fuwhelper = True
    if helper.get_file_exists(cip, device.me_paths, '\\Windows\\FUWhelper.dll'):
        transfer_fuwhelper = False
        device.me_paths.fuwhelper_file = '\\Windows\\FUWhelper.dll'
        device.log.append(f'Firmware upgrade helper already present on terminal.')
    else:
        if helper.get_file_exists(cip, device.me_paths, device.me_paths.fuwhelper_file):
            transfer_fuwhelper = False
            device.log.append(f'Firmware upgrade helper already present on terminal.')

    # Download firmware upgrade wizard helper to terminal
    if transfer_fuwhelper:
        try:
            fuw_helper_file = types.MEFile('FUWhelper.dll',
                                        True,
                                        True,
                                        firmware_helper_path)
            resp = download_file(cip, device, fuw_helper_file, '\\Storage Card', progress)
            if not(resp):
                device.log.append(f'Failed to upgrade terminal.')
                return False
            
            time.sleep(10)
        except Exception as e:
            device.log.append(f'Exception: {str(e)}')
            device.log.append(f'Traceback: {traceback.format_exc()}')
            device.log.append(f'Failed to upgrade terminal.')
            return False            

    # Prepare terminal for firmware upgrade card
    try:
        if not(fuwhelper.get_folder_exists(cip, device.me_paths, '\\Storage Card')):
            fuwhelper.create_folder(cip, device.me_paths, '\\Storage Card')
        if not(fuwhelper.get_folder_exists(cip, device.me_paths, '\\Storage Card\\vfs')):
            fuwhelper.create_folder(cip, device.me_paths, '\\Storage Card\\vfs')
        if not(fuwhelper.get_folder_exists(cip, device.me_paths, '\\Storage Card\\vfs\\platform firmware')):
            fuwhelper.create_folder(cip, device.me_paths, '\\Storage Card\\vfs\\platform firmware')
        if (fuwhelper.get_file_exists(cip, device.me_paths, '\\Storage Card\\Step2.dat')):
            fuwhelper.delete_file(cip, device.me_paths, '\\Storage Card\\Step2.dat')
        if fuwhelper.get_process_running(cip, device.me_paths, 'MERuntime.exe'):
            fuwhelper.stop_process_me(cip, device.me_paths)

        fuwhelper.get_file_exists(cip, device.me_paths, '\\Windows\\useroptions.txt')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Traceback: {traceback.format_exc()}')
        device.log.append(f'Failed to upgrade terminal.')
        return False

    # Download firmware upgrade card to terminal
    try:
        fuw_image_file = types.MEFile('SC.IMG',
                                    True,
                                    True,
                                    firmware_image_path)
        resp = download_file(cip, device, fuw_image_file, '\\vfs\\platform firmware', progress)
        if not(resp):
            device.log.append(f'Failed to upgrade terminal.')
            return False
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Traceback: {traceback.format_exc()}')
        device.log.append(f'Failed to upgrade terminal.')
        return False

    return True

def upload(cip: comms.Driver, device: types.MEDeviceInfo, rem_file_path: str, progress: Optional[Callable[[str, int, int], None]] = None) -> bytearray:
    # Create a transfer instance on the terminal
    try:
        transfer_instance, total_bytes = files.create_transfer_instance_upload(cip, f'{rem_file_path}')
        device.log.append(f'Create transfer instance {transfer_instance} for upload.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create transfer instance for upload')
        return False

    # Transfer file chunk by chunk
    try:
        resp_binary = files.execute_transfer_upload(cip, transfer_instance, total_bytes, progress)
        device.log.append(f'Uploaded {rem_file_path} using transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to upload {rem_file_path} using transfer instance {transfer_instance}.')

    # Delete transfer instance on the terminal
    try:
        files.delete_transfer_instance(cip, transfer_instance)
        device.log.append(f'Deleted transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to delete transfer instance {transfer_instance}.')

    return resp_binary
    
def upload_file(cip: comms.Driver, device: types.MEDeviceInfo, local_file_path: str, rem_file_path: str, progress: Optional[Callable[[str, int, int], None]] = None):
    resp_binary = upload(cip, device, rem_file_path, progress)
    with open(local_file_path, 'wb') as dest_file:
        dest_file.write(resp_binary)

def upload_list(cip: comms.Driver, transfer_instance: int, rem_file_path: str) -> list[str]:
    resp_binary = upload(cip, transfer_instance, rem_file_path)
    resp_str = "".join([chr(b) for b in resp_binary if b != 0])
    resp_list = resp_str.split(':')
    return resp_list

def upload_mer_file(cip: comms.Driver, device: types.MEDeviceInfo, file: types.MEFile, rem_file: types.MEFile, progress: Optional[Callable[[str, int, int], None]] = None) -> bool:
    # Verify file exists on terminal
    try:
        if helper.get_file_exists_mer(cip, device.me_paths, rem_file.name):
            device.log.append(f'File {rem_file.name} exists on terminal.')
        else:
            device.log.append(f'File {rem_file.name} does not exist on terminal.')
            return False
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to check if file {rem_file.name} exists on terminal.')
        return False

    # Perform upload
    try:
        rem_file_path = f'{device.me_paths.runtime}\\{rem_file.name}'
        upload_file(cip, device, file.path, rem_file_path, progress)
        device.log.append(f'Uploaded {rem_file_path} to {file.path}.')
    except:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to upload {rem_file_path} to {file.path}.')
        return False

    return True

def upload_med_list(cip: comms.Driver, device: types.MEDeviceInfo) -> list[str]:
    # Create list on the terminal
    try:
        helper.create_file_list_med(cip, device.me_paths)
        device.log.append(f'Created *.MED list on terminal.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create *.MED list on terminal.')
        return None

    # Perform upload
    file_list = None
    try:
        rem_file_path = f'{device.me_paths.upload_list}'
        file_list = upload_list(cip, device, rem_file_path)
        device.log.append(f'Uploaded {rem_file_path}.')
    except:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to upload {rem_file_path}.')
        return None

    # Delete list on the terminal
    try:
        helper.delete_file_list(cip, device.me_paths)
        device.log.append(f'Deleted *.MED list on terminal.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to delete *.MED list on terminal.')

    return file_list

def upload_mer_list(cip: comms.Driver, device: types.MEDeviceInfo) -> list[str]:
    # Create *.MER list
    try:
        helper.create_file_list_mer(cip, device.me_paths)
        device.log.append(f'Created *.MER list on terminal.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create *.MER list on terminal.')
        return None

    # Perform upload
    file_list = None
    try:
        rem_file_path = f'{device.me_paths.upload_list}'
        file_list = upload_list(cip, device, rem_file_path)
        device.files = file_list
        device.log.append(f'Uploaded {rem_file_path}.')
    except:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to upload {rem_file_path}.')
        return None

    # Delete list on the terminal
    try:
        helper.delete_file_list(cip, device.me_paths)
        device.log.append(f'Deleted *.MER list on terminal.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to delete *.MER list on terminal.')

    return file_list

def reboot(cip: comms.Driver, device: types.MEDeviceInfo):
    cip1 = comms.Driver(cip._original_path)
    cip1.timeout = 0.25
    cip1.open()
    try:
        # Execute reboot
        device.log.append(f'Rebooting terminal.')
        helper.reboot(cip1, device.me_paths)

        # If we made it here... the reboot function didn't throw
        # an exception, which means it didn't reboot.
        #
        # Further investigation needed.
        device.log.append(f'Initial reboot unsuccessful.')

        # Currently tailoring additional action to terminals where:
        #
        # Firmware version is 12.108+
        # Application is set to load at startup
        # ME Station is set to run application at startup
        # Current App is defined
        if ((device.me_identity.major_rev < 12) or ((device.me_identity.major_rev == 12) and (device.me_identity.minor_rev < 108))):
            device.log.append(f'Did not attempt additional reboot because terminal is below minimum applicable version.')
            return
        if (registry.get_startup_options(cip) != 1):
            device.log.append(f'Did not attempt additional reboot because terminal Startup Options are not set to Run Current Application.')
            return
        startup_file = os.path.basename(registry.get_startup_mer(cip).replace('\\','/'))
        if not(startup_file.lower().endswith('.mer')):
            device.log.append(f'Did not attempt additional reboot because terminal does not have valid startup *.MER defined.')
            return

        # Rebuild existing ME Startup Shortcut on terminal
        delete_logs = registry.get_startup_delete_logs(cip)
        replace_comms = registry.get_startup_replace_comms(cip)
        device.log.append(f'Setting file: {startup_file} to run at startup with Replace Comms: {replace_comms}, Delete Logs: {delete_logs}.')
        helper.create_me_shortcut(cip, device.me_paths, startup_file, replace_comms, delete_logs)

        # Execute reboot
        device.log.append(f'Rebooting terminal.')
        helper.reboot(cip1, device.me_paths)
    except Exception as e:
        # Unlike most CIP messages, this one is expected to
        # create an exception.  When it is received by the terminal,
        # the device reboots and breaks the socket.
        if (str(e) != 'failed to receive reply'): raise e
    
    cip1.close()